import logging
from concurrent.futures import ProcessPoolExecutor
from functools import lru_cache
from typing import Callable, Dict, List, Optional, Tuple

import torch

from lhotse import CutSet, FeatureExtractor
from lhotse.cut import compute_supervisions_frame_mask
from lhotse.dataset.collation import collate_audio, collate_features, collate_matrices, collate_vectors, \
    read_audio_from_cuts
from lhotse.utils import LOG_EPSILON, compute_num_frames, ifnone, supervision_to_frames, supervision_to_samples


class BatchIO:
    """
    Converts a :class:`CutSet` into a collated batch of audio representations.
    These representations can be e.g. audio samples or features.
    They might also be single or multi channel.

    All InputStrategies support the ``executor`` parameter in the constructor.
    It allows to pass a ``ThreadPoolExecutor`` or a ``ProcessPoolExecutor``
    to parallelize reading audio/features from wherever they are stored.
    Note that this approach is incompatible with specifying the ``num_workers``
    to ``torch.utils.data.DataLoader``, but in some instances may be faster.

    .. note:: This is a base class that only defines the interface.

    .. automethod:: __call__
    """
    def __init__(self, num_workers: int = 0) -> None:
        self.num_workers = num_workers

    def __call__(self, cuts: CutSet) -> Tuple[torch.Tensor, torch.IntTensor]:
        """Returns a tensor with collated input signals, and a tensor of length of each signal before padding."""
        raise NotImplementedError()

    def supervision_intervals(self, cuts: CutSet) -> Dict[str, torch.Tensor]:
        """
        Returns a dict that specifies the start and end bounds for each supervision,
        as a 1-D int tensor.

        Depending on the strategy, the dict should look like:

        .. code-block:

            {
                "sequence_idx": tensor(shape=(S,)),
                "start_frame": tensor(shape=(S,)),
                "num_frames": tensor(shape=(S,)),
            }

        or

        .. code-block:

            {
                "sequence_idx": tensor(shape=(S,)),
                "start_sample": tensor(shape=(S,)),
                "num_samples": tensor(shape=(S,))
            }

        Where ``S`` is the total number of supervisions encountered in the :class:`CutSet`.
        Note that ``S`` might be different than the number of cuts (``B``).
        ``sequence_idx`` means the index of the corresponding feature matrix (or cut) in a batch.
        """
        raise NotImplementedError()

    def supervision_masks(self, cuts: CutSet) -> torch.Tensor:
        """
        Returns a collated batch of masks, marking the supervised regions in cuts.
        They are zero-padded to the longest cut.

        Depending on the strategy implementation, it is expected to be a
        tensor of shape ``(B, NF)`` or ``(B, NS)``, where ``B`` denotes the number of cuts,
        ``NF`` the number of frames and ``NS`` the total number of samples.
        ``NF`` and ``NS`` are determined by the longest cut in a batch.
        """
        raise NotImplementedError()


class PrecomputedFeatures(BatchIO):
    """
    :class:`InputStrategy` that reads pre-computed features, whose manifests
    are attached to cuts, from disk.

    It pads the feature matrices, if needed.

    .. automethod:: __call__
    """

    def __call__(self, cuts: CutSet) -> Tuple[torch.Tensor, torch.IntTensor]:
        """
        Reads the pre-computed features from disk/other storage.
        The returned shape is ``(B, T, F) => (batch_size, num_frames, num_features)``.

        :return: a tensor with collated features, and a tensor of ``num_frames`` of each cut before padding."""
        return collate_features(cuts, executor=_get_executor(self.num_workers))

    def supervision_intervals(self, cuts: CutSet) -> Dict[str, torch.Tensor]:
        """
        Returns a dict that specifies the start and end bounds for each supervision,
        as a 1-D int tensor, in terms of frames:

        .. code-block:

            {
                "sequence_idx": tensor(shape=(S,)),
                "start_frame": tensor(shape=(S,)),
                "num_frames": tensor(shape=(S,))
            }

        Where ``S`` is the total number of supervisions encountered in the :class:`CutSet`.
        Note that ``S`` might be different than the number of cuts (``B``).
        ``sequence_idx`` means the index of the corresponding feature matrix (or cut) in a batch.
        """
        start_frames, nums_frames = zip(*(
            supervision_to_frames(sup, cut.frame_shift, cut.sampling_rate, max_frames=cut.num_frames)
            for cut in cuts
            for sup in cut.supervisions
        ))
        sequence_idx = [i for i, c in enumerate(cuts) for s in c.supervisions]
        return {
            'sequence_idx': torch.tensor(sequence_idx, dtype=torch.int32),
            'start_frame': torch.tensor(start_frames, dtype=torch.int32),
            'num_frames': torch.tensor(nums_frames, dtype=torch.int32)
        }

    def supervision_masks(self, cuts: CutSet, use_alignment_if_exists: Optional[str] = None) -> torch.Tensor:
        """Returns the mask for supervised frames.
        :param use_alignment_if_exists: optional str, key for alignment type to use for generating the mask. If not
            exists, fall back on supervision time spans.
        """
        return collate_vectors([cut.supervisions_feature_mask(use_alignment_if_exists=use_alignment_if_exists) for cut in cuts])


class AudioSamples(BatchIO):
    """
    :class:`InputStrategy` that reads single-channel recordings, whose manifests
    are attached to cuts, from disk (or other audio source).

    It pads the recordings, if needed.

    .. automethod:: __call__
    """
    def __call__(self, cuts: CutSet) -> Tuple[torch.Tensor, torch.IntTensor]:
        """
        Reads the audio samples from recordings on disk/other storage.
        The returned shape is ``(B, T) => (batch_size, num_samples)``.

        :return: a tensor with collated audio samples, and a tensor of ``num_samples`` of each cut before padding.
        """
        return collate_audio(cuts, executor=_get_executor(self.num_workers))

    def supervision_intervals(self, cuts: CutSet) -> Dict[str, torch.Tensor]:
        """
        Returns a dict that specifies the start and end bounds for each supervision,
        as a 1-D int tensor, in terms of samples:

        .. code-block:

            {
                "sequence_idx": tensor(shape=(S,)),
                "start_sample": tensor(shape=(S,)),
                "num_samples": tensor(shape=(S,))
            }

        Where ``S`` is the total number of supervisions encountered in the :class:`CutSet`.
        Note that ``S`` might be different than the number of cuts (``B``).
        ``sequence_idx`` means the index of the corresponding feature matrix (or cut) in a batch.

        """
        start_samples, nums_samples = zip(*(
            supervision_to_samples(sup, cut.sampling_rate)
            for cut in cuts
            for sup in cut.supervisions
        ))
        sequence_idx = [i for i, c in enumerate(cuts) for s in c.supervisions]
        return {
            'sequence_idx': torch.tensor(sequence_idx, dtype=torch.int32),
            'start_sample': torch.tensor(start_samples, dtype=torch.int32),
            'num_samples': torch.tensor(nums_samples, dtype=torch.int32)
        }

    def supervision_masks(self, cuts: CutSet, use_alignment_if_exists: Optional[str] = None) -> torch.Tensor:
        """Returns the mask for supervised samples.
        :param use_alignment_if_exists: optional str, key for alignment type to use for generating the mask. If not
            exists, fall back on supervision time spans.
        """
        return collate_vectors([cut.supervisions_audio_mask(use_alignment_if_exists=use_alignment_if_exists) for cut in cuts])


class OnTheFlyFeatures(BatchIO):
    """
    :class:`InputStrategy` that reads single-channel recordings, whose manifests
    are attached to cuts, from disk (or other audio source).
    Then, it uses a :class:`FeatureExtractor` to compute their features on-the-fly.

    It pads the recordings, if needed.

    .. note:
        The batch feature extraction performed here is not as efficient as it could be,
        but it allows to use arbitrary feature extraction method that may work on
        a single recording at a time.

    .. automethod:: __call__
    """

    def __init__(
            self,
            extractor: FeatureExtractor,
            wave_transforms: List[Callable[[torch.Tensor], torch.Tensor]] = None,
            num_workers: int = 0,
    ) -> None:
        """
        OnTheFlyFeatures' constructor.

        :param extractor: the feature extractor used on-the-fly (individually on each waveform).
        :param wave_transforms: an optional list of transforms applied on the batch of audio
            waveforms collated into a single tensor, right before the feature extraction.
        """
        super().__init__(num_workers=num_workers)
        self.extractor = extractor
        self.wave_transforms = ifnone(wave_transforms, [])

    def __call__(self, cuts: CutSet) -> Tuple[torch.Tensor, torch.IntTensor]:
        """
        Reads the audio samples from recordings on disk/other storage
        and computes their features.
        The returned shape is ``(B, T, F) => (batch_size, num_frames, num_features)``.

        :return: a tensor with collated features, and a tensor of ``num_frames`` of each cut before padding.
        """
        audios = read_audio_from_cuts(cuts, executor=_get_executor(self.num_workers))

        for tfnm in self.wave_transforms:
            for idx in range(len(audios)):
                audios[idx] = tfnm(audios[idx])

        features_single = []
        for idx, cut in enumerate(cuts):
            samples = audios[idx].numpy()
            try:
                features = self.extractor.extract(samples, cuts[idx].sampling_rate)
            except:
                logging.error(f"Error while extracting the features for cut with ID {cut.id} -- details:\n{cut}")
                raise
            features_single.append(torch.from_numpy(features))
        features_batch = collate_matrices(features_single, padding_value=LOG_EPSILON)

        feature_lens = torch.tensor([
            compute_num_frames(
                cut.duration,
                self.extractor.frame_shift,
                cut.sampling_rate
            ) for cut in cuts
        ], dtype=torch.int32)

        return features_batch, feature_lens

    def supervision_intervals(self, cuts: CutSet) -> Dict[str, torch.Tensor]:
        """
        Returns a dict that specifies the start and end bounds for each supervision,
        as a 1-D int tensor, in terms of frames:

        .. code-block:

            {
                "sequence_idx": tensor(shape=(S,)),
                "start_frame": tensor(shape=(S,)),
                "num_frames": tensor(shape=(S,))
            }

        Where ``S`` is the total number of supervisions encountered in the :class:`CutSet`.
        Note that ``S`` might be different than the number of cuts (``B``).
        ``sequence_idx`` means the index of the corresponding feature matrix (or cut) in a batch.
        """
        start_frames, nums_frames = zip(*(
            supervision_to_frames(sup, self.extractor.frame_shift, cut.sampling_rate)
            for cut in cuts
            for sup in cut.supervisions
        ))
        sequence_idx = [i for i, c in enumerate(cuts) for s in c.supervisions]
        return {
            'sequence_idx': torch.tensor(sequence_idx, dtype=torch.int32),
            'start_frame': torch.tensor(start_frames, dtype=torch.int32),
            'num_frames': torch.tensor(nums_frames, dtype=torch.int32)
        }

    def supervision_masks(self, cuts: CutSet, use_alignment_if_exists: Optional[str] = None) -> torch.Tensor:
        """Returns the mask for supervised samples.
        :param use_alignment_if_exists: optional str, key for alignment type to use for generating the mask. If not
            exists, fall back on supervision time spans.
        """
        return collate_vectors(
            [
                compute_supervisions_frame_mask(
                    cut,
                    frame_shift=self.extractor.frame_shift,
                    use_alignment_if_exists=use_alignment_if_exists
                ) for cut in cuts
            ]
        )


@lru_cache(maxsize=1)
def _get_executor(max_workers: int = 0) -> Optional[ProcessPoolExecutor]:
    """
    This function caches a process pool in the global state of a given process.
    It's useful for keeping a process pool alive across different invocations within the
    same process for efficiency.
    We intend it to be used for efficient data reads withing a task executed in a
    parent process pool.
    """
    if max_workers <= 0:
        return None
    return ProcessPoolExecutor(max_workers=max_workers)





import click

from typing import Optional

from lhotse.bin.modes import prepare
from lhotse.recipes.cslu_kids import prepare_cslu_kids
from lhotse.utils import Pathlike

__all__ = ["cslu_kids"]


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "--absolute-paths",
    type=bool,
    default=True,
    help="Use absolute paths for recordings",
)
@click.option(
    "--normalize-text",
    type=bool,
    default=True,
    help="Remove noise tags (<bn>, <bs>) from spontaneous speech transcripts",
)
def cslu_kids(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
    absolute_paths: Optional[bool] = False,
    normalize_text: Optional[bool] = True,
):
    """CSLU Kids corpus data preparation."""
    prepare_cslu_kids(
        corpus_dir,
        output_dir=output_dir,
        absolute_paths=absolute_paths,
        normalize_text=normalize_text,
    )



from pathlib import Path

import click

from lhotse.bin.modes.cli_base import cli
from lhotse.utils import Pathlike


@cli.group()
def kaldi():
    """Kaldi import/export related commands."""
    pass


@kaldi.command(name='import')
@click.argument('data_dir', type=click.Path(exists=True, file_okay=False))
@click.argument('sampling_rate', type=int)
@click.argument('manifest_dir', type=click.Path())
@click.option('-f', '--frame-shift', type=float,
              help='Frame shift (in seconds) is required to support reading feats.scp.')
def import_(data_dir: Pathlike, sampling_rate: int, manifest_dir: Pathlike, frame_shift: float):
    """
    Convert a Kaldi data dir DATA_DIR into a directory MANIFEST_DIR of lhotse manifests. Ignores feats.scp.
    The SAMPLING_RATE has to be explicitly specified as it is not available to read from DATA_DIR.
    """
    from lhotse.kaldi import load_kaldi_data_dir

    recording_set, maybe_supervision_set, maybe_feature_set = load_kaldi_data_dir(
        path=data_dir,
        sampling_rate=sampling_rate,
        frame_shift=frame_shift
    )
    manifest_dir = Path(manifest_dir)
    manifest_dir.mkdir(parents=True, exist_ok=True)
    recording_set.to_file(manifest_dir / 'recordings.jsonl.gz')
    if maybe_supervision_set is not None:
        maybe_supervision_set.to_file(manifest_dir / 'supervisions.jsonl.gz')
    if maybe_feature_set is not None:
        maybe_feature_set.to_file(manifest_dir / 'features.jsonl.gz')


@kaldi.command()
@click.argument('recordings', type=click.Path(exists=True, dir_okay=False))
@click.argument('supervisions', type=click.Path(exists=True, dir_okay=False))
@click.argument('output_dir', type=click.Path())
def export(recordings: Pathlike, supervisions: Pathlike, output_dir: Pathlike):
    """
    Convert a pair of ``RecordingSet`` and ``SupervisionSet`` manifests into a Kaldi-style data directory.
    """
    from lhotse import load_manifest
    from lhotse.kaldi import export_to_kaldi

    output_dir = Path(output_dir)
    export_to_kaldi(
        recordings=load_manifest(recordings),
        supervisions=load_manifest(supervisions),
        output_dir=output_dir
    )
    click.secho('Export completed! You likely need to run the following Kaldi commands:', bold=True, fg='yellow')
    click.secho(f'  utils/utt2spk_to_spk2utt.pl {output_dir}/utt2spk > {output_dir}/spk2utt', fg='yellow')
    click.secho(f'  utils/fix_data_dir.sh {output_dir}', fg='yellow')



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.ami import download_ami, prepare_ami
from lhotse.utils import Pathlike

__all__ = ['ami']


@prepare.command(context_settings=dict(show_default=True))
@click.argument('corpus_dir', type=click.Path(exists=True, dir_okay=True))
@click.argument('output_dir', type=click.Path())
@click.option(
    '--annotations',
    type=click.Path(),
    default=None,
    help="Provide if annotations are download in a different directory than corpus.",
)
@click.option(
    '--mic',
    type=click.Choice(['ihm', 'ihm-mix', 'sdm', 'mdm'], case_sensitive=False),
    default='ihm',
    help='AMI microphone setting.',
)
@click.option(
    '--partition',
    type=click.Choice(
        ['scenario-only', 'full-corpus', 'full-corpus-asr'], case_sensitive=False
    ),
    default='full-corpus-asr',
    help='Data partition to use (see http://groups.inf.ed.ac.uk/ami/corpus/datasets.shtml).',
)
@click.option(
    '--max-pause',
    type=float,
    default=0.0,
    help='Max pause allowed between word segments to combine segments.',
)
def ami(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
    annotations: Pathlike,
    mic: str,
    partition: str,
    max_pause: float,
):
    """AMI data preparation."""
    prepare_ami(
        corpus_dir,
        annotations_dir=annotations,
        output_dir=output_dir,
        mic=mic,
        partition=partition,
        max_pause=max_pause,
    )


@download.command(context_settings=dict(show_default=True))
@click.argument('target_dir', type=click.Path())
@click.option(
    '--annotations',
    type=click.Path(),
    default=None,
    help="To download annotations in a different directory than corpus.",
)
@click.option(
    '--mic',
    type=click.Choice(['ihm', 'ihm-mix', 'sdm', 'mdm'], case_sensitive=False),
    default='ihm',
    help='AMI microphone setting.',
)
@click.option(
    '--url',
    type=str,
    default='http://groups.inf.ed.ac.uk/ami',
    help='AMI data downloading URL.',
)
@click.option(
    '--force-download',
    type=bool,
    default=False,
    help='If True, download even if file is present.',
)
def ami(
    target_dir: Pathlike,
    annotations: Pathlike,
    mic: str,
    url: str,
    force_download: bool,
):
    """AMI download."""
    download_ami(
        target_dir,
        annotations_dir=annotations,
        mic=mic,
        url=url,
        force_download=force_download,
    )



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.musan import download_musan, prepare_musan
from lhotse.utils import Pathlike

__all__ = ['musan']


@prepare.command()
@click.argument('corpus_dir', type=click.Path(exists=True, dir_okay=True))
@click.argument('output_dir', type=click.Path())
@click.option('--use-vocals/--no-vocals', default=True, help='Whether to include vocal music in "music" part.')
def musan(
        corpus_dir: Pathlike,
        output_dir: Pathlike,
        use_vocals: bool
):
    """MUSAN data preparation."""
    prepare_musan(corpus_dir, output_dir=output_dir, use_vocals=use_vocals)


@download.command()
@click.argument('target_dir', type=click.Path())
def musan(
        target_dir: Pathlike
):
    """MUSAN download."""
    download_musan(target_dir)



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes import download_vctk, prepare_vctk
from lhotse.utils import Pathlike

__all__ = ['vctk']


@prepare.command()
@click.argument('corpus_dir', type=click.Path(exists=True, dir_okay=True))
@click.argument('output_dir', type=click.Path())
def vctk(
        corpus_dir: Pathlike,
        output_dir: Pathlike
):
    """VCTK data preparation."""
    prepare_vctk(corpus_dir, output_dir=output_dir)


@download.command()
@click.argument('target_dir', type=click.Path())
def vctk(
        target_dir: Pathlike
):
    """VCTK download."""
    download_vctk(target_dir)



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes import download_libritts, prepare_libritts
from lhotse.utils import Pathlike

__all__ = ['libritts']


@prepare.command(context_settings=dict(show_default=True))
@click.argument('corpus_dir', type=click.Path(exists=True, dir_okay=True))
@click.argument('output_dir', type=click.Path())
@click.option('-j', '--num-jobs', type=int, default=1,
              help='How many jobs to use (can give good speed-ups with slow disks).')
def libritts(
        corpus_dir: Pathlike,
        output_dir: Pathlike,
        num_jobs: int
):
    """LibriTTs data preparation."""
    prepare_libritts(corpus_dir, output_dir=output_dir, num_jobs=num_jobs)


@download.command(context_settings=dict(show_default=True))
@click.argument('target_dir', type=click.Path())
def libritts(
        target_dir: Pathlike,
):
    """LibriTTS data download."""
    download_libritts(target_dir)



from typing import Optional, Sequence, Union

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.mtedx import download_mtedx, prepare_mtedx
from lhotse.utils import Pathlike

__all__ = ['mtedx']


@prepare.command(context_settings=dict(show_default=True))
@click.argument('corpus_dir', type=click.Path(exists=True, dir_okay=True))
@click.argument('output_dir', type=click.Path())
@click.option('-j', '--num-jobs', type=int, default=1,
              help='How many threads to use (can give good speed-ups with slow disks).')
@click.option('-l', '--lang', multiple=True, default=['all'],
    help='Specify which languages to prepare, e.g., '
    '        lhoste prepare librispeech mtedx_corpus data -l de -l fr -l es '
)
def mtedx(
        corpus_dir: Pathlike,
        output_dir: Pathlike,
        num_jobs: int,
        lang: Optional[Union[str, Sequence[str]]],
):
    """MTEDx ASR data preparation."""
    prepare_mtedx(corpus_dir, output_dir=output_dir,
        num_jobs=num_jobs, languages=lang
    )


@download.command(context_settings=dict(show_default=True))
@click.argument('target_dir', type=click.Path())
@click.option('-l', '--lang', multiple=True, default=['all'],
    help='Specify which languages to download, e.g., '
    '        lhoste download mtedx . -l de -l fr -l es '
    '        lhoste download mtedx'
)
def mtedx(
        target_dir: Pathlike,
        lang: Optional[Union[str, Sequence[str]]],
):
    """MTEDx download."""
    download_mtedx(target_dir, languages=lang)
